import os
import re
import shutil
from io import TextIOWrapper
from typing import Tuple

import numpy as np
import pandas as pd
import pyemc

from pmd.util import HiddenPrints, Pmdlogging, build_dir

PSP_FORCE_FIELD_OPTIONS = ('opls-lbcc', 'opls-cm1a', 'gaff2-gasteiger',
                           'gaff2-am1bcc')
EMC_FORCE_FIELD_OPTIONS = ('pcff', 'opls-aa')

EMC_EXTS = ('esh', 'data', 'in', 'params', 'vmd', 'emc.gz', 'pdb.gz', 'psf.gz')
EMC_COEFF_EXCLUSIONS = ('bb', 'ba', 'mbt', 'ebt', 'at', 'aat', 'bb13', 'aa')


class Builder:

    def __init__(self, force_field: str,
                 force_field_options: Tuple[str]) -> None:
        self._force_field = force_field
        self._validate_force_field(force_field_options)

    def __repr__(self) -> str:
        return type(self).__name__

    def _validate_force_field(self, options):
        if self._force_field not in options:
            raise ValueError(f'Invalid {self} force_field, valid options are '
                             f'{", ".join(options)}')

    def write_data(self) -> None:
        raise NotImplementedError

    def write_functional_form(self) -> None:
        raise NotImplementedError


class EMC(Builder):
    '''Object to perform system structure generation using
    (EMC)[http://montecarlo.sourceforge.net/]: Enhanced Monte Carlo package.
    This object should be used as input argument of `System` or `Lammps`
    objects

    Attributes:
        force_field (str): Force field, options are `"pcff"` and `"opls-aa"`
    '''

    def __init__(self, force_field: str) -> None:
        super().__init__(force_field, EMC_FORCE_FIELD_OPTIONS)

    @staticmethod
    def _remove_brackets_around_asterisks(smiles):
        smiles = smiles.replace('[*]', '*')
        return smiles

    @build_dir
    def write_data(self, output_dir: str, smiles: str, density: float,
                   natoms_total: int, length: int, nchains: int,
                   data_fname: str, cleanup: bool) -> None:

        smiles = self._remove_brackets_around_asterisks(smiles)

        previous_dir = os.getcwd()
        os.chdir(output_dir)

        tmp_file_prefilx = 'system'
        # Write .esh file required to run EMC
        with open(f'{tmp_file_prefilx}.esh', 'w') as f:
            f.write('# EMC input file generated by PMD\n')
            f.write('ITEM OPTIONS\n')
            f.write('replace true\n')
            f.write(f'field {self._force_field}\n')
            f.write(f'density {density}\n')
            f.write(f'ntotal {natoms_total}\n')
            f.write('ITEM END\n')
            f.write('\n')
            f.write('ITEM GROUPS\n')
            f.write(f'RU {smiles},1,RU:2\n')
            f.write('terminator *[H],1,RU:1,1,RU:2\n')
            f.write('ITEM END\n')
            f.write('\n')
            f.write('ITEM CLUSTERS\n')
            f.write('poly alternate 1\n')
            f.write('ITEM END\n')
            f.write('\n')
            f.write('ITEM POLYMERS\n')
            f.write('poly\n')
            f.write(f'1 RU,{length},terminator,2\n')
            f.write('ITEM END\n')

        try:
            Pmdlogging.info('Launching EMC...')
            pyemc.setup(f'{tmp_file_prefilx}.esh')
            pyemc.build('build.emc')

            # store parameters from .params file
            params = {}
            key = None
            with open(f'{tmp_file_prefilx}.params', 'r') as lines:
                for line in lines:
                    parts = line.split()
                    fw = parts[0] if len(parts) > 0 else ''
                    # put an empty list when reaching a Coeffs section
                    if line.startswith('#') and line.endswith('Coeffs\n'):
                        key = line.lstrip('# ')
                        params[key] = []
                    elif '_coeff' in fw and key:
                        params[key].append(line.lstrip(fw))

            # store data from .data file
            final_file_before_coeffs = []
            final_file_after_coeffs = []
            before = True
            with open(f'{tmp_file_prefilx}.data', 'r') as lines:
                for line in lines:
                    if line == 'Atoms\n':
                        before = False
                    if before:
                        final_file_before_coeffs.append(line)
                    else:
                        final_file_after_coeffs.append(line)

            # There are double empty lines at the end of EMC data file
            final_file_after_coeffs = final_file_after_coeffs[:-1]

            # combine data and parameters into the final data file
            with open(data_fname, 'w') as f:
                for line in final_file_before_coeffs:
                    f.write(line)
                for param, param_lines in params.items():
                    f.write(param)
                    f.write('\n')
                    for line in param_lines:
                        for coeff in EMC_COEFF_EXCLUSIONS:
                            line = line.replace(f' {coeff} ', ' ')
                        f.write(line)
                    f.write('\n')
                for line in final_file_after_coeffs:
                    f.write(line)

            Pmdlogging.info(f'System file - {data_fname} '
                            f'successfully created in {output_dir}')

        finally:
            # Clean up all EMC generated files except for the data file
            if cleanup:
                fnames = ['build.emc']
                for ext in EMC_EXTS:
                    fnames.append(f'{tmp_file_prefilx}.{ext}')
                for fname in fnames:
                    try:
                        os.remove(fname)
                    except Exception:
                        print(f'problem removing {fname} during cleanup')

            os.chdir(previous_dir)

    def write_functional_form(self, f: TextIOWrapper) -> None:
        if self._force_field == 'pcff':
            f.write(f'{"pair_style":<15} lj/class2/coul/long 9.5 9.5\n')
            f.write(f'{"pair_modify":<15} mix sixthpower tail yes\n')
            f.write(f'{"kspace_style":<15} pppm/cg 1e-4\n')
            f.write(f'{"bond_style":<15} class2\n')
            f.write(f'{"angle_style":<15} class2\n')
            f.write(f'{"dihedral_style":<15} class2\n')
            f.write(f'{"improper_style":<15} class2\n')
            f.write(f'{"special_bonds":<15} lj/coul 0 0 1\n')

        elif self._force_field == 'opls-aa':
            f.write(f'{"pair_style":<15} lj/cut/coul/long 9.5 9.5\n')
            f.write(f'{"pair_modify":<15} mix geometric tail yes\n')
            f.write(f'{"kspace_style":<15} pppm/cg 1e-4\n')
            f.write(f'{"bond_style":<15} harmonic\n')
            f.write(f'{"angle_style":<15} harmonic\n')
            f.write(f'{"dihedral_style":<15} multi/harmonic\n')
            f.write(f'{"improper_style":<15} harmonic\n')
            f.write(f'{"special_bonds":<15} lj/coul 0.0 0.0 0.5\n')


class PSP(Builder):
    '''Object to perform system structure generation using
    (PSP)[https://github.com/Ramprasad-Group/PSP]: Polymer Structure Predictor
    package. This object should be used as input argument of `System` or
    `Lammps` objects

    Attributes:
        force_field (str): Force field, options are `"opls-lbcc"`,
        `"opls-cm1a"`, `"gaff2-gasteiger"`, `"gaff2-am1bcc"`
    '''

    def __init__(self, force_field: str) -> None:
        super().__init__(force_field, PSP_FORCE_FIELD_OPTIONS)

    @staticmethod
    def _add_brackets_to_asterisks(smiles):
        stars_no_bracket = re.findall(r'(?<!\[)\*(?!\])', smiles)
        if len(stars_no_bracket) == 2:
            smiles = smiles.replace('*', '[*]')
        return smiles

    def _is_opls_force_field(self):
        return self._force_field.startswith('opls')

    def _run_psp(self, input_data: dict, density: float, data_fname: str,
                 output_dir: str, cleanup: bool) -> None:
        try:
            import psp.AmorphousBuilder as ab
        except ImportError:
            raise ImportError('Please install PSP to use PSP builder')

        Pmdlogging.info(
            'PSP builder: Creating the system, this may take a while...')
        try:
            with HiddenPrints():
                amor = ab.Builder(pd.DataFrame(data=input_data),
                                  density=density,
                                  OutDir=output_dir)
                amor.Build()

                if self._is_opls_force_field:
                    amor.get_opls(
                        output_fname=data_fname,
                        lbcc_charges=self._force_field.endswith('opls-lbcc'))
                else:
                    amor.get_gaff2(
                        output_fname=data_fname,
                        atom_typing='antechamber',
                        am1bcc_charges=self._force_field.endswith('am1bcc'),
                        swap_dict={
                            'ns': 'n',
                            'nt': 'n',
                            'nv': 'nh'
                        })
            Pmdlogging.info(
                f'System file - {data_fname} successfully created in {output_dir}'
            )
        finally:
            if cleanup:
                force_field_dname = [
                    'ligpargen'
                ] if self._is_opls_force_field else ['pysimm']
                dnames = ['molecules', 'packmol'] + force_field_dname
                for dir in dnames:
                    try:
                        shutil.rmtree(os.path.join(output_dir, dir))
                    except FileNotFoundError:
                        pass

                fnames = ['amor_model.data', 'amor_model.vasp']
                for file in fnames:
                    try:
                        os.remove(os.path.join(output_dir, file))
                    except FileNotFoundError:
                        pass

                fnames = ['output_MB.csv', 'molecules.csv']
                for file in fnames:
                    try:
                        os.remove(file)
                    except FileNotFoundError:
                        pass

    def write_data(self, output_dir: str, smiles: str, density: float,
                   natoms_total: int, length: int, nchains: int,
                   data_fname: str, cleanup: bool) -> None:

        smiles = self._add_brackets_to_asterisks(smiles)
        input_data = {
            'ID': ['Poly'],
            'smiles': [smiles],
            'Len': [length],
            'Num': [nchains],
            'NumConf': [1],
            'Loop': [False],
            'LeftCap': ['[*][H]'],
            'RightCap': ['[*][H]']
        }
        self._run_psp(input_data, density, data_fname, output_dir, cleanup)

    def write_solvent_data(self, output_dir, smiles, solvent_smiles, density,
                           natoms_total, length, nsolvents, nchains,
                           data_fname, cleanup):

        smiles = self._add_brackets_to_asterisks(smiles)
        input_data = {
            'ID': ['Sol', 'Poly'],
            'smiles': [solvent_smiles, smiles],
            'Len': [1, length],
            'Num': [nsolvents, nchains],
            'NumConf': [1, 1],
            'Loop': [False, False],
            'LeftCap': [np.nan, '[*][H]'],
            'RightCap': [np.nan, '[*][H]']
        }
        self._run_psp(input_data, density, data_fname, output_dir, cleanup)

    def write_functional_form(self, f: TextIOWrapper) -> None:

        if self._force_field.startswith('opls'):
            f.write(f'{"pair_style":<15} lj/cut/coul/long 9.0\n')
            f.write(f'{"pair_modify":<15} mix geometric tail yes\n')
            f.write(f'{"kspace_style":<15} pppm 1e-4\n')
            f.write(f'{"bond_style":<15} harmonic\n')
            f.write(f'{"angle_style":<15} harmonic\n')
            f.write(f'{"dihedral_style":<15} opls\n')
            f.write(f'{"improper_style":<15} cvff\n')
            f.write(f'{"special_bonds":<15} lj/coul 0.0 0.0 0.5\n')

        elif self._force_field.startswith('gaff2'):
            f.write(f'{"pair_style":<15} lj/cut/coul/long 12.0 12.0\n')
            f.write(f'{"pair_modify":<15} mix arithmetic\n')
            f.write(f'{"kspace_style":<15} pppm 1e-4\n')
            f.write(f'{"bond_style":<15} harmonic\n')
            f.write(f'{"angle_style":<15} harmonic\n')
            f.write(f'{"dihedral_style":<15} fourier\n')
            f.write(f'{"improper_style":<15} cvff\n')
            f.write(f'{"special_bonds":<15} amber\n')
